import threading
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Callable, Dict, List, Optional

import ingenialogger

try:
    from fsoe_master.fsoe_master import (
        ApplicationParameter,
        Dictionary,
        DictionaryItem,
        DictionaryItemInput,
        DictionaryItemInputOutput,
        MasterHandler,
        StateData,
    )

    if TYPE_CHECKING:
        from fsoe_master.fsoe_master import State

except ImportError:
    FSOE_MASTER_INSTALLED = False
else:
    FSOE_MASTER_INSTALLED = True

from ingenialink.enums.register import REG_ACCESS, REG_DTYPE
from ingenialink.ethercat.register import EthercatRegister
from ingenialink.pdo import RPDOMap, RPDOMapItem, TPDOMap, TPDOMapItem
from ingenialink.utils._utils import dtype_value

from ingeniamotion.enums import FSoEState
from ingeniamotion.exceptions import IMTimeoutError
from ingeniamotion.metaclass import DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController


@dataclass
class FSoEError:
    """FSoE Error descriptor"""

    servo: str
    transition_name: str
    description: str


class FSoEMasterHandler:
    """FSoE Master Handler.

    Args:
        slave_address: The servo's FSoE address.
        connection_id: The FSoE connection ID.
        watchdog_timeout: The FSoE master watchdog timeout in seconds.

    """

    STO_COMMAND_KEY = 0x040
    STO_COMMAND_UID = "STO_COMMAND"
    SS1_COMMAND_KEY = 0x050
    SS1_COMMAND_UID = "SS1_COMMAND"
    SAFE_INPUTS_KEY = 0x070
    SAFE_INPUTS_UID = "SAFE_INPUTS"
    PROCESS_DATA_COMMAND = 0x36

    def __init__(
        self,
        slave_address: int,
        connection_id: int,
        watchdog_timeout: float,
        application_parameters: List["ApplicationParameter"],
        report_error_callback: Callable[[str, str], None],
    ):
        if not FSOE_MASTER_INSTALLED:
            return
        self.__master_handler = MasterHandler(
            dictionary=self._saco_phase_1_dictionary(),
            slave_address=slave_address,
            connection_id=connection_id,
            watchdog_timeout_s=watchdog_timeout,
            application_parameters=application_parameters,
            report_error_callback=report_error_callback,
            state_change_callback=self.__state_change_callback,
        )
        self._configure_master()
        self.__safety_master_pdu = RPDOMap()
        self.__safety_slave_pdu = TPDOMap()
        self._configure_pdo_maps()
        self.__running = False
        self.__state_is_data = threading.Event()

        # The saco slave might take a while to answer with a valid command
        # During it's initialization it will respond with 0's, that are ignored
        # To avoid triggering additional errors
        self.__in_initial_reset = False

    def _start(self) -> None:
        """Start the FSoE Master handler."""
        self.__in_initial_reset = True
        self.__master_handler.start()
        self.__running = True

    def stop(self) -> None:
        """Stop the master handler"""
        self.__master_handler.stop()
        self.__in_initial_reset = False
        self.__running = False

    def delete(self) -> None:
        """Delete the master handler"""
        self.__master_handler.delete()

    def _configure_pdo_maps(self) -> None:
        """Configure the PDOMaps used for the Safety PDUs."""
        PDUMapper.configure_rpdo_map(self.safety_master_pdu_map)
        PDUMapper.configure_tpdo_map(self.safety_slave_pdu_map)

    def _configure_master(self) -> None:
        """Configure the FSoE master handler."""
        self._map_outputs()
        self._map_inputs()

    def _map_outputs(self) -> None:
        """Configure the FSoE master handler's SafeOutputs."""
        # Phase 1 mapping
        self.__master_handler.master.dictionary_map.add_by_key(self.STO_COMMAND_KEY, bits=1)
        self.__master_handler.master.dictionary_map.add_by_key(self.SS1_COMMAND_KEY, bits=1)
        self.__master_handler.master.dictionary_map.add_padding(bits=7)
        self.__master_handler.master.dictionary_map.add_padding(bits=7)

    def _map_inputs(self) -> None:
        """Configure the FSoE master handler's SafeInputs."""
        # Phase 1 mapping
        self.__master_handler.slave.dictionary_map.add_by_key(self.STO_COMMAND_KEY, bits=1)
        self.__master_handler.slave.dictionary_map.add_by_key(self.SS1_COMMAND_KEY, bits=1)
        self.__master_handler.slave.dictionary_map.add_padding(bits=7)
        self.__master_handler.slave.dictionary_map.add_by_key(self.SAFE_INPUTS_KEY, bits=1)
        self.__master_handler.slave.dictionary_map.add_padding(bits=6)

    def get_request(self) -> None:
        """Set the FSoE master handler request to the Safety Master PDU PDOMap"""
        if not self.__running:
            self._start()
        self.safety_master_pdu_map.set_item_bytes(self.__master_handler.get_request())

    def set_reply(self) -> None:
        """Get the FSoE slave response from the Safety Slave PDU PDOMap and set it
        to the FSoE master handler."""
        reply = self.safety_slave_pdu_map.get_item_bytes()
        if self.__in_initial_reset:
            if reply[0] == 0:
                # Byte 0 of FSoE frame should always be the command
                # 0 is not a valid command
                return
            else:
                self.__in_initial_reset = False

        self.__master_handler.set_reply(reply)

    def sto_deactivate(self) -> None:
        """Set the STO command to deactivate the STO"""
        self.__master_handler.set_fail_safe(False)
        self.__master_handler.dictionary.set(self.STO_COMMAND_UID, True)

    def sto_activate(self) -> None:
        """Set the STO command to activate the STO"""
        self.__master_handler.dictionary.set(self.STO_COMMAND_UID, False)

    def ss1_deactivate(self) -> None:
        """Set the SS1 command to deactivate the SS1"""
        self.__master_handler.set_fail_safe(False)
        self.__master_handler.dictionary.set(self.SS1_COMMAND_UID, True)

    def ss1_activate(self) -> None:
        """Set the SS1 command to activate the SS1"""
        self.__master_handler.dictionary.set(self.SS1_COMMAND_UID, False)

    def safe_inputs_value(self) -> bool:
        """Get the safe inputs register value"""
        safe_inputs_value = self.__master_handler.dictionary.get(self.SAFE_INPUTS_UID)
        if not isinstance(safe_inputs_value, bool):
            raise ValueError(f"Wrong value type. Expected type bool, got {type(safe_inputs_value)}")
        return safe_inputs_value

    def is_sto_active(self) -> bool:
        """Check the STO state.

        Returns:
            True if the STO is active. False otherwise.

        """
        sto_command = self.__master_handler.dictionary.get(self.STO_COMMAND_UID)
        if not isinstance(sto_command, bool):
            raise ValueError(f"Wrong value type. Expected type bool, got {type(sto_command)}")
        return sto_command

    def __state_change_callback(self, state: "State") -> None:
        if state == StateData:
            self.__state_is_data.set()
        else:
            self.__state_is_data.clear()

    def wait_for_data_state(self, timeout: Optional[float] = None) -> None:
        """Wait the FSoE master handler to reach the Data state.

        Args:
            timeout : how many seconds to wait for the FSoE master to reach the
                Data state, if ``None`` it will wait forever.
                ``None`` by default.

        Raises:
            IMTimeoutError: If the Data state is not reached within the timeout.

        """
        if self.__state_is_data.wait(timeout=timeout) is False:
            raise IMTimeoutError("The FSoE Master did not reach the Data state")

    def _saco_phase_1_dictionary(self) -> "Dictionary":
        """Get the SaCo phase 1 dictionary instance"""
        sto_command_dict_item = DictionaryItemInputOutput(
            key=self.STO_COMMAND_KEY,
            name=self.STO_COMMAND_UID,
            data_type=DictionaryItem.DataTypes.BOOL,
            fail_safe_input_value=True,
        )
        ss1_command_dict_item = DictionaryItemInputOutput(
            key=self.SS1_COMMAND_KEY,
            name=self.SS1_COMMAND_UID,
            data_type=DictionaryItem.DataTypes.BOOL,
            fail_safe_input_value=True,
        )
        safe_input_dict_item = DictionaryItemInput(
            key=self.SAFE_INPUTS_KEY,
            name=self.SAFE_INPUTS_UID,
            data_type=DictionaryItem.DataTypes.BOOL,
            fail_safe_value=False,
        )
        return Dictionary(
            [
                sto_command_dict_item,
                ss1_command_dict_item,
                safe_input_dict_item,
            ]
        )

    @property
    def safety_master_pdu_map(self) -> RPDOMap:
        """The PDOMap used for the Safety Master PDU."""
        return self.__safety_master_pdu

    @property
    def safety_slave_pdu_map(self) -> TPDOMap:
        """The PDOMap used for the Safety Slave PDU."""
        return self.__safety_slave_pdu

    @property
    def state(self) -> FSoEState:
        """Get the FSoE master state."""
        return FSoEState(self.__master_handler.state.id)

    @property
    def running(self) -> bool:
        """True if FSoE Master is started, else False"""
        return self.__running


class PDUMapper:
    """Helper class to configure the Safety PDU PDOMaps."""

    FSOE_RPDO_MAP_1_INDEX = 0x1700
    FSOE_TPDO_MAP_1_INDEX = 0x1B00

    # Phase 1 mapping
    FSOE_COMMAND_SIZE_BITS = 8
    STO_COMMAND_SIZE_BITS = 1
    SS1_COMMAND_SIZE_BITS = 1
    STO_COMMAND_PADDING_SIZE_BITS = 6
    SBC_COMMAND_SIZE_BITS = 1
    SBC_COMMAND_PADDING_SIZE_BITS = 7
    CRC_O_SIZE_BITS = 16
    CONN_ID_SIZE_BITS = 16
    SAFE_INPUTS_SIZE_BITS = 1
    SAFE_INPUTS_PADDING_SIZE_BITS = 6

    @classmethod
    def configure_rpdo_map(cls, rpdo_map: RPDOMap) -> None:
        """Configure the RPDOMap used for the Safety Master PDU.

        Args:
            rpdo_map: The RPDOMap instance.

        """
        rpdo_map.map_register_index = cls.FSOE_RPDO_MAP_1_INDEX
        fsoe_command_item = RPDOMapItem(size_bits=cls.FSOE_COMMAND_SIZE_BITS)
        rpdo_map.add_item(fsoe_command_item)
        sto_command_item = RPDOMapItem(size_bits=cls.STO_COMMAND_SIZE_BITS)
        rpdo_map.add_item(sto_command_item)
        ss1_command_item = RPDOMapItem(size_bits=cls.SS1_COMMAND_SIZE_BITS)
        rpdo_map.add_item(ss1_command_item)
        sto_padding_item = RPDOMapItem(size_bits=cls.STO_COMMAND_PADDING_SIZE_BITS)
        rpdo_map.add_item(sto_padding_item)
        sbc_command_item = RPDOMapItem(size_bits=cls.SBC_COMMAND_SIZE_BITS)
        rpdo_map.add_item(sbc_command_item)
        sbc_padding_item = RPDOMapItem(size_bits=cls.SBC_COMMAND_PADDING_SIZE_BITS)
        rpdo_map.add_item(sbc_padding_item)
        crc_0_item = RPDOMapItem(size_bits=cls.CRC_O_SIZE_BITS)
        rpdo_map.add_item(crc_0_item)
        conn_id_item = RPDOMapItem(size_bits=cls.CONN_ID_SIZE_BITS)
        rpdo_map.add_item(conn_id_item)

    @classmethod
    def configure_tpdo_map(cls, tpdo_map: TPDOMap) -> None:
        """Configure the TPDOMap used for the Safety Slave PDU.

        Args:
            tpdo_map: The TPDOMap instance.

        """
        tpdo_map.map_register_index = cls.FSOE_TPDO_MAP_1_INDEX
        fsoe_command_item = TPDOMapItem(size_bits=cls.FSOE_COMMAND_SIZE_BITS)
        tpdo_map.add_item(fsoe_command_item)
        sto_command_item = TPDOMapItem(size_bits=cls.STO_COMMAND_SIZE_BITS)
        tpdo_map.add_item(sto_command_item)
        ss1_command_item = TPDOMapItem(size_bits=cls.SS1_COMMAND_SIZE_BITS)
        tpdo_map.add_item(ss1_command_item)
        sto_padding_item = TPDOMapItem(size_bits=cls.STO_COMMAND_PADDING_SIZE_BITS)
        tpdo_map.add_item(sto_padding_item)
        sbc_command_item = TPDOMapItem(size_bits=cls.SBC_COMMAND_SIZE_BITS)
        tpdo_map.add_item(sbc_command_item)
        safe_inputs_item = TPDOMapItem(size_bits=cls.SAFE_INPUTS_SIZE_BITS)
        tpdo_map.add_item(safe_inputs_item)
        safe_inputs_padding_item = TPDOMapItem(size_bits=cls.SAFE_INPUTS_PADDING_SIZE_BITS)
        tpdo_map.add_item(safe_inputs_padding_item)
        crc_0_item = TPDOMapItem(size_bits=cls.CRC_O_SIZE_BITS)
        tpdo_map.add_item(crc_0_item)
        conn_id_item = TPDOMapItem(size_bits=cls.CONN_ID_SIZE_BITS)
        tpdo_map.add_item(conn_id_item)


class FSoEMaster:
    """FSoE Master.

    Args:
        motion_controller: The MotionController instance.

    """

    DEFAULT_WATCHDOG_TIMEOUT_S = 1

    SAFETY_ADDRESS_REGISTER = EthercatRegister(
        idx=0x4193, subidx=0x00, dtype=REG_DTYPE.U16, access=REG_ACCESS.RW
    )
    SAFE_INPUTS_MAP_REGISTER = EthercatRegister(
        identifier="SAFE_INPUTS_MAP",
        idx=0x46D2,
        subidx=0x00,
        dtype=REG_DTYPE.U16,
        access=REG_ACCESS.RW,
    )
    SS1_TIME_TO_STO_REGISTER = EthercatRegister(
        identifier="SS1_TIME_TO_STO",
        idx=0x6651,
        subidx=0x01,
        dtype=REG_DTYPE.U16,
        access=REG_ACCESS.RW,
    )

    def __init__(self, motion_controller: "MotionController") -> None:
        self.logger = ingenialogger.get_logger(__name__)
        self.__mc = motion_controller
        self.__handlers: Dict[str, FSoEMasterHandler] = {}
        self.__next_connection_id = 1
        self._error_observers: List[Callable[[FSoEError], None]] = []
        self.__fsoe_configured = False

    def create_fsoe_master_handler(
        self,
        servo: str = DEFAULT_SERVO,
        fsoe_master_watchdog_timeout: float = DEFAULT_WATCHDOG_TIMEOUT_S,
    ) -> None:
        """Create an FSoE Master handler linked to a Safe servo drive.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            fsoe_master_watchdog_timeout: The FSoE master watchdog timeout in seconds.

        """
        slave_address = self.get_safety_address(servo)
        application_parameters = self._get_application_parameters(servo)
        master_handler = FSoEMasterHandler(
            slave_address,
            self.__next_connection_id,
            fsoe_master_watchdog_timeout,
            application_parameters,
            partial(self._notify_errors, servo=servo),
        )
        self.__handlers[servo] = master_handler
        self.__next_connection_id += 1

    def configure_pdos(self, start_pdos: bool = False) -> None:
        """Configure the PDOs used for the Safety PDUs.

        Args:
            start_pdos: if ``True``, start the PDO exchange, if ``False``
                the PDO exchange should be started after. ``False`` by default.

        """
        self._set_pdo_maps_to_slaves()
        self._subscribe_to_pdo_thread_events()
        if start_pdos:
            self.__mc.capture.pdo.start_pdos()
        self.__fsoe_configured = True

    def stop_master(self, stop_pdos: bool = False) -> None:
        """Stop all the FSoE Master handlers.

        Args:
            stop_pdos: if ``True``, stop the PDO exchange. ``False`` by default.

        """
        for master_handler in self.__handlers.values():
            if master_handler.running:
                master_handler.stop()
        if self.__fsoe_configured:
            self._unsubscribe_from_pdo_thread_events()
            self._remove_pdo_maps_from_slaves()
        else:
            self.logger.warning("FSoE master is already stopped")
        self.__fsoe_configured = False
        if stop_pdos:
            self.__mc.capture.pdo.stop_pdos()

    def sto_deactivate(self, servo: str = DEFAULT_SERVO) -> None:
        """Deactivate the Safety Torque Off.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self.__handlers[servo]
        master_handler.sto_deactivate()

    def sto_activate(self, servo: str = DEFAULT_SERVO) -> None:
        """Activate the Safety Torque Off.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self.__handlers[servo]
        master_handler.sto_activate()

    def ss1_deactivate(self, servo: str = DEFAULT_SERVO) -> None:
        """Deactivate the SS1.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self.__handlers[servo]
        master_handler.ss1_deactivate()

    def ss1_activate(self, servo: str = DEFAULT_SERVO) -> None:
        """Activate the SS1.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self.__handlers[servo]
        master_handler.ss1_activate()

    def get_safety_inputs_value(self, servo: str = DEFAULT_SERVO) -> bool:
        """Get a drive's safe inputs register value.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
           The safe inputs value.

        """
        master_handler = self.__handlers[servo]
        return master_handler.safe_inputs_value()

    def get_safety_address(self, servo: str = DEFAULT_SERVO) -> int:
        """Get the drive's FSoE slave address.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
            The FSoE slave address.

        """
        drive = self.__mc.servos[servo]
        value = drive.read(self.SAFETY_ADDRESS_REGISTER)
        if not isinstance(value, int):
            raise ValueError(f"Wrong safety address value type. Expected int, got {type(value)}")
        return value

    def set_safety_address(self, address: int, servo: str = DEFAULT_SERVO) -> None:
        """Set the drive's FSoE slave address.

        Args:
            address: The address to be set.
            servo: servo alias to reference it. ``default`` by default.

        """
        drive = self.__mc.servos[servo]
        drive.write(self.SAFETY_ADDRESS_REGISTER, data=address)

    def check_sto_active(self, servo: str = DEFAULT_SERVO) -> bool:
        """Check if the STO is active in a given servo.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
            True if the STO is active. False otherwise.

        """
        master_handler = self.__handlers[servo]
        return master_handler.is_sto_active()

    def wait_for_state_data(
        self, servo: str = DEFAULT_SERVO, timeout: Optional[float] = None
    ) -> None:
        """Wait for an FSoE master handler to reach the Data state.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            timeout : how many seconds to wait for the FSoE master to reach the
                Data state, if ``None`` it will wait forever.
                ``None`` by default.

        """
        master_handler = self.__handlers[servo]
        master_handler.wait_for_data_state(timeout)

    def get_fsoe_master_state(self, servo: str = DEFAULT_SERVO) -> FSoEState:
        """Get the servo's FSoE master handler state.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
            The servo's FSoE master handler state.

        """
        master_handler = self.__handlers[servo]
        return master_handler.state

    def subscribe_to_errors(self, callback: Callable[[FSoEError], None]) -> None:
        """Subscribe to the FSoE errors.

        Args:
            callback: Subscribed callback function.

        """
        if callback in self._error_observers:
            return
        self._error_observers.append(callback)

    def unsubscribe_from_errors(self, callback: Callable[[FSoEError], None]) -> None:
        """Unsubscribe from the FSoE errors.

        Args:
            callback: Subscribed callback function.

        """
        if callback not in self._error_observers:
            return
        self._error_observers.remove(callback)

    def _notify_errors(self, transition_name: str, error_description: str, servo: str) -> None:
        """Notify subscribers when an FSoE error occurs.

        Args:
            transition_name: FSoE transition name.
            error_description: FSoE error description.
            servo: The servo alias.

        """
        for callback in self._error_observers:
            callback(FSoEError(servo, transition_name, error_description))

    def _delete_master_handler(self, servo: str = DEFAULT_SERVO) -> None:
        """Delete the master handler instance

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        if servo not in self.__handlers:
            return
        self.__handlers[servo].delete()
        del self.__handlers[servo]

    def _subscribe_to_pdo_thread_events(self) -> None:
        """Subscribe to the PDO thread events.

        This allows to send the Safety Master PDU and to retrieve the Safety Slave PDU.

        """
        self.__mc.capture.pdo.subscribe_to_send_process_data(self._get_request)
        self.__mc.capture.pdo.subscribe_to_receive_process_data(self._set_reply)
        self.__mc.capture.pdo.subscribe_to_exceptions(self._pdo_thread_exception_handler)

    def _unsubscribe_from_pdo_thread_events(self) -> None:
        """Unsubscribe from the PDO thread events."""
        self.__mc.capture.pdo.unsubscribe_to_send_process_data(self._get_request)
        self.__mc.capture.pdo.unsubscribe_to_receive_process_data(self._set_reply)
        self.__mc.capture.pdo.unsubscribe_to_exceptions(self._pdo_thread_exception_handler)

    def _set_pdo_maps_to_slaves(self) -> None:
        """Set the PDOMaps to be used by the Safety PDUs to the slaves."""
        for servo, master_handler in self.__handlers.items():
            rpdo_map = master_handler.safety_master_pdu_map
            tpdo_map = master_handler.safety_slave_pdu_map
            self.__mc.capture.pdo.set_pdo_maps_to_slave(rpdo_map, tpdo_map, servo)

    def _remove_pdo_maps_from_slaves(self) -> None:
        """Remove the PDOMaps used by the Safety PDUs from the slaves."""
        for servo, master_handler in self.__handlers.items():
            self.__mc.capture.pdo.remove_rpdo_map(servo, master_handler.safety_master_pdu_map)
            self.__mc.capture.pdo.remove_tpdo_map(servo, master_handler.safety_slave_pdu_map)

    def _get_request(self) -> None:
        """Callback method to send the FSoE Master handlers requests to the
        corresponding FSoE slave."""
        for master_handler in self.__handlers.values():
            master_handler.get_request()

    def _set_reply(self) -> None:
        """Callback method to provide the FSoE Slaves responses to their
        corresponding FSoE Master handler."""
        for master_handler in self.__handlers.values():
            master_handler.set_reply()

    def _pdo_thread_exception_handler(self, exc: Exception) -> None:
        """Callback method for the PDO thread exceptions."""
        self.logger.error(
            "The FSoE Master lost connection to the FSoE slaves. "
            f"An exception occurred during the PDO exchange: {exc}"
        )

    def _get_application_parameters(self, servo: str) -> List["ApplicationParameter"]:
        """Get values of the application parameters"""
        drive = self.__mc.servos[servo]
        application_parameters = []
        for register in [
            self.SAFE_INPUTS_MAP_REGISTER,
            self.SS1_TIME_TO_STO_REGISTER,
        ]:
            register_size_bytes, _ = dtype_value[register.dtype]
            application_parameter = ApplicationParameter(
                name=register.identifier,
                initial_value=drive.read(register),
                n_bytes=register_size_bytes,
            )
            application_parameters.append(application_parameter)
        return application_parameters



import importlib
import time
from typing import Dict

import numpy as np
import pytest
import rpyc
from ingenialink.canopen.network import CAN_BAUDRATE, CAN_DEVICE
from ping3 import ping
from virtual_drive.core import VirtualDrive

from ingeniamotion import MotionController
from ingeniamotion.enums import SensorType

from .setups.descriptors import (
    DriveCanOpenSetup,
    DriveEcatSetup,
    DriveEthernetSetup,
    DriveHwSetup,
    EthercatMultiSlaveSetup,
    Setup,
    VirtualDriveSetup,
)
from .setups.environment_control import (
    ManualUserEnvironmentController,
    RackServiceEnvironmentController,
    VirtualDriveEnvironmentController,
)

test_report_key = pytest.StashKey[Dict[str, pytest.CollectReport]]()


def pytest_addoption(parser):
    parser.addoption(
        "--setup",
        action="store",
        default="tests.setups.tests_setup.TESTS_SETUP",
        help="Module and location from which to import the setup."
        "It will default to a file that you can create on"
        "tests_setup.py inside of the folder setups with a variable called TESTS_SETUP"
        "This variable must define, or must be assigned to a Setup instance",
    )
    parser.addoption(
        "--job_name",
        action="store",
        default="ingeniamotion - Unknown",
        help="Name of the executing job. Will be set to rack service to have more info of the logs",
    )


@pytest.fixture(scope="session")
def tests_setup(request) -> Setup:
    # Get option from argument and split by dots (modules and last variable name)
    setup_location = request.config.getoption("--setup").split(".")
    # Dynamically import the python module
    setup_module = importlib.import_module(".".join(setup_location[:-1]))
    # Get the variable by variable name
    setup = getattr(setup_module, setup_location[-1])
    return setup


def connect_ethernet(mc, config, alias):
    mc.communication.connect_servo_ethernet(config.ip, config.dictionary, alias=alias)


def connect_soem(mc, config: DriveEcatSetup, alias):
    mc.communication.connect_servo_ethercat(
        config.ifname,
        config.slave,
        config.dictionary,
        alias,
    )


def connect_canopen(mc, config: DriveCanOpenSetup, alias):
    device = CAN_DEVICE(config.device)
    baudrate = CAN_BAUDRATE(config.baudrate)
    mc.communication.connect_servo_canopen(
        device,
        config.dictionary,
        config.node_id,
        baudrate,
        config.channel,
        alias=alias,
    )


@pytest.fixture(scope="session")
def motion_controller(tests_setup: Setup, pytestconfig, request):
    alias = "test"
    mc = MotionController()

    if isinstance(tests_setup, DriveHwSetup):
        if tests_setup.use_rack_service:
            rack_service_client = request.getfixturevalue("connect_to_rack_service")
            drive_idx, drive = tests_setup.get_rack_drive(rack_service_client)
            environment = RackServiceEnvironmentController(
                rack_service_client, default_drive_idx=drive_idx
            )
        else:
            environment = ManualUserEnvironmentController(pytestconfig)

        if isinstance(tests_setup, DriveEcatSetup):
            connect_soem(mc, tests_setup, alias)
        elif isinstance(tests_setup, DriveCanOpenSetup):
            connect_canopen(mc, tests_setup, alias)
        elif isinstance(tests_setup, DriveEthernetSetup):
            connect_ethernet(mc, tests_setup, alias)
        else:
            raise NotImplementedError

        if tests_setup.config_file is not None:
            mc.configuration.restore_configuration(servo=alias)
            mc.configuration.load_configuration(tests_setup.config_file, servo=alias)
        yield mc, alias, environment
        environment.reset()
        mc.communication.disconnect(alias)

    elif isinstance(tests_setup, EthercatMultiSlaveSetup):
        if tests_setup.drives[0].use_rack_service:
            environment = RackServiceEnvironmentController(
                request.getfixturevalue("connect_to_rack_service")
            )
        else:
            environment = ManualUserEnvironmentController(pytestconfig)

        aliases = []
        for drive in tests_setup.drives:
            mc.communication.connect_servo_ethercat(
                interface_name=drive.ifname,
                slave_id=drive.slave,
                dict_path=drive.dictionary,
                alias=drive.identifier,
            )
            aliases.append(drive.identifier)

        yield mc, aliases, environment
        environment.reset()
    elif isinstance(tests_setup, VirtualDriveSetup):
        virtual_drive = VirtualDrive(tests_setup.port, tests_setup.dictionary)
        virtual_drive.start()
        connect_ethernet(mc, tests_setup, alias)
        environment = VirtualDriveEnvironmentController(virtual_drive.environment)

        yield mc, alias, environment

        environment.reset()
        virtual_drive.stop()
    else:
        raise NotImplementedError


@pytest.fixture(autouse=True)
def disable_motor_fixture(pytestconfig, motion_controller, tests_setup):
    yield

    if isinstance(tests_setup, DriveHwSetup):
        mc, alias, environment = motion_controller
        mc.motion.motor_disable(servo=alias)
        mc.motion.fault_reset(servo=alias)


@pytest.fixture
def motion_controller_teardown(motion_controller, pytestconfig, tests_setup: Setup):
    yield motion_controller
    if isinstance(tests_setup, DriveHwSetup):
        mc, alias, environment = motion_controller
        mc.motion.motor_disable(servo=alias)
        if tests_setup.config_file is not None:
            mc.configuration.load_configuration(tests_setup.config_file, servo=alias)
        mc.motion.fault_reset(servo=alias)


@pytest.fixture
def disable_monitoring_disturbance(motion_controller):
    yield
    mc, alias, environment = motion_controller
    mc.capture.clean_monitoring_disturbance(servo=alias)


@pytest.fixture(scope="session")
def feedback_list(motion_controller):
    mc, alias, environment = motion_controller
    fdbk_lst = [
        mc.configuration.get_commutation_feedback(servo=alias),
        mc.configuration.get_reference_feedback(servo=alias),
        mc.configuration.get_velocity_feedback(servo=alias),
        mc.configuration.get_position_feedback(servo=alias),
        mc.configuration.get_auxiliar_feedback(servo=alias),
    ]
    return set(fdbk_lst)


@pytest.fixture
def clean_and_restore_feedbacks(motion_controller):
    mc, alias, environment = motion_controller
    comm = mc.configuration.get_commutation_feedback(servo=alias)
    ref = mc.configuration.get_reference_feedback(servo=alias)
    vel = mc.configuration.get_velocity_feedback(servo=alias)
    pos = mc.configuration.get_position_feedback(servo=alias)
    aux = mc.configuration.get_auxiliar_feedback(servo=alias)
    mc.configuration.set_commutation_feedback(SensorType.INTGEN, servo=alias)
    mc.configuration.set_reference_feedback(SensorType.INTGEN, servo=alias)
    mc.configuration.set_velocity_feedback(SensorType.INTGEN, servo=alias)
    mc.configuration.set_position_feedback(SensorType.INTGEN, servo=alias)
    mc.configuration.set_auxiliar_feedback(SensorType.QEI, servo=alias)
    yield
    mc.configuration.set_commutation_feedback(comm, servo=alias)
    mc.configuration.set_reference_feedback(ref, servo=alias)
    mc.configuration.set_velocity_feedback(vel, servo=alias)
    mc.configuration.set_position_feedback(pos, servo=alias)
    mc.configuration.set_auxiliar_feedback(aux, servo=alias)


@pytest.fixture()
def skip_if_monitoring_not_available(motion_controller):
    mc, alias, environment = motion_controller
    try:
        mc.capture._check_version(alias)
    except NotImplementedError:
        pytest.skip("Monitoring is not available")


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # store test results for each phase of a call, which can be "setup", "call", "teardown"
    item.stash.setdefault(test_report_key, {})[rep.when] = rep


@pytest.fixture(scope="function", autouse=True)
def load_configuration_if_test_fails(pytestconfig, request, motion_controller, tests_setup: Setup):
    mc, alias, environment = motion_controller
    yield

    report = request.node.stash[test_report_key]

    if isinstance(tests_setup, DriveHwSetup) and (
        report["setup"].failed or ("call" not in report) or report["call"].failed
    ):
        if tests_setup.config_file is not None:
            mc.configuration.load_configuration(tests_setup.config_file, servo=alias)
        mc.motion.fault_reset(servo=alias)


def mean_actual_velocity_position(mc, servo, velocity=False, n_samples=200, sampling_period=0):
    samples = np.zeros(n_samples)
    get_actual_value_dict = {
        True: mc.motion.get_actual_velocity,
        False: mc.motion.get_actual_position,
    }
    for sample_idx in range(n_samples):
        value = get_actual_value_dict[velocity](servo=servo)
        samples[sample_idx] = value
        time.sleep(sampling_period)
    return np.mean(samples)


@pytest.fixture(scope="module", autouse=True)
def load_configuration_after_each_module(pytestconfig, motion_controller, tests_setup: Setup):
    yield motion_controller

    if isinstance(tests_setup, DriveHwSetup):
        mc, alias, environment = motion_controller
        mc.motion.motor_disable(servo=alias)
        if tests_setup.config_file is not None:
            mc.configuration.load_configuration(tests_setup.config_file, servo=alias)


@pytest.fixture(scope="session")
def connect_to_rack_service(request):
    rack_service_port = 33810
    client = rpyc.connect("localhost", rack_service_port, config={"sync_request_timeout": None})
    client.root.set_job_name(request.config.getoption("--job_name"))
    yield client.root
    client.close()


@pytest.fixture(scope="session", autouse=True)
def load_firmware(pytestconfig, tests_setup: Setup, request):
    if not isinstance(tests_setup, DriveHwSetup):
        return

    if not tests_setup.use_rack_service:
        return

    client = request.getfixturevalue("connect_to_rack_service")
    number_of_drives = len(client.exposed_get_configuration().drives)

    # Reboot drive
    client.exposed_turn_off_ps()
    time.sleep(1)
    client.exposed_turn_on_ps()

    # Wait for all drives to turn-on, for 90 seconds
    timeout = 90
    wait_until = time.time() + timeout
    mc = MotionController()
    while True:
        if time.time() >= wait_until:
            raise TimeoutError(f"Could not find drives in {timeout} after rebooting")

        if isinstance(tests_setup, DriveEcatSetup):
            n_found = len(mc.communication.scan_servos_ethercat(tests_setup.ifname))
            if n_found == number_of_drives:
                break
        elif isinstance(tests_setup, DriveCanOpenSetup):
            # Temporal workaround
            # Canopen transceiver setup generates BUS-off errors when scanning servos
            # Until the transceiver is not changed or a better method is implemented on rack service
            # it will wait for some time and assume they are connected
            time.sleep(60)
            break
        elif isinstance(tests_setup, DriveEthernetSetup):
            ping_result = ping(dest_addr=tests_setup.ip)
            # The response delay in seconds/milliseconds, False on error and None on timeout.
            if isinstance(ping_result, float):
                break
        else:
            raise NotImplementedError

    # Load firmware (if necessary, if it's already loaded it will do nothing)
    drive_idx, drive = tests_setup.get_rack_drive(client)
    client.exposed_firmware_load(
        drive_idx, tests_setup.fw_file, drive.product_code, drive.serial_number
    )



from typing import TYPE_CHECKING, Union

import ingenialogger

from ingeniamotion.enums import GPI, GPO, DigitalVoltageLevel, GPIOPolarity
from ingeniamotion.exceptions import IMException
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO, MCMetaClass

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController


class InputsOutputs(metaclass=MCMetaClass):
    GPIO_IN_VALUE_REGISTER = "IO_IN_VALUE"
    GPIO_IN_POLARITY_REGISTER = "IO_IN_POLARITY"

    GPIO_OUT_VALUE_REGISTER = "IO_OUT_VALUE"
    GPIO_OUT_SET_POINT_REGISTER = "IO_OUT_SET_POINT"
    GPIO_OUT_POLARITY_REGISTER = "IO_OUT_POLARITY"

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self.logger = ingenialogger.get_logger(__name__)

    @staticmethod
    def __get_gpio_bit_value(io_register_value: int, gpio_id: Union[GPI, GPO]) -> bool:
        """Get the bit value of a specific GPIO.

        Args:
            io_register_value: Register value including the bits of all GPIOs.
            gpio_id: The GPIO identifier.

        Returns:
            GPIO bit value.
        """
        gpio_bit_mask = 1 << (gpio_id - 1)
        return bool(io_register_value & gpio_bit_mask)

    @staticmethod
    def __set_gpio_bit_value(
        io_register_value: int, gpio_id: Union[GPI, GPO], bit_value: bool
    ) -> int:
        """Set the bit value of a specific GPIO.

        Args:
            io_register_value: Register value including the bits of all GPIOs.
            gpio_id: The GPIO identifier.
            bit_value: Bit value to be set.

        Returns:
            The register value with the corresponding bit set accordingly.
        """
        gpio_bit_mask = 1 << (gpio_id - 1)
        if bit_value:
            new_register_value = io_register_value | gpio_bit_mask
        else:
            new_register_value = io_register_value & ~gpio_bit_mask

        return new_register_value

    def get_gpi_polarity(
        self, gpi_id: GPI, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> GPIOPolarity:
        """Get the polarity of a single GPI.

        Args:
            gpi_id: the GPI to get the polarity.
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        Returns:
            Polarity of the specified GPI.

        """
        gpi_polarity = int(
            self.mc.communication.get_register(
                self.GPIO_IN_POLARITY_REGISTER, servo=servo, axis=axis
            )
        )
        gpi_bit_polarity = self.__get_gpio_bit_value(gpi_polarity, gpi_id)
        return GPIOPolarity(int(gpi_bit_polarity))

    def set_gpi_polarity(
        self,
        gpi_id: GPI,
        polarity: GPIOPolarity,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set the polarity of a single GPI.

        Args:
            gpi_id: the GPI polarity to be set.
            polarity: polarity to be set.
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        """
        gpis_actual_polarity = int(
            self.mc.communication.get_register(
                self.GPIO_IN_POLARITY_REGISTER, servo=servo, axis=axis
            )
        )
        gpis_new_polarity = self.__set_gpio_bit_value(
            gpis_actual_polarity, gpi_id, bool(polarity.value)
        )
        self.mc.communication.set_register(
            self.GPIO_IN_POLARITY_REGISTER, gpis_new_polarity, servo=servo, axis=axis
        )

    def get_gpi_voltage_level(
        self, gpi_id: GPI, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> DigitalVoltageLevel:
        """Get the board voltage level (not the logic level at the uC) of a single GPI.

        .. warning::
            Ensure the polarity is right, if not this function will return a wrong value.

        Args:
            gpi_id: the gpi to get the voltage level
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        Returns:
            LOW if the voltage level is 0, HIGH if the voltage level is 1.

        """
        gpi_value = int(
            self.mc.communication.get_register(self.GPIO_IN_VALUE_REGISTER, servo=servo, axis=axis)
        )
        gpi_bit_value = self.__get_gpio_bit_value(gpi_value, gpi_id)

        return DigitalVoltageLevel(gpi_bit_value)

    def get_gpo_polarity(
        self, gpo_id: GPO, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> GPIOPolarity:
        """Get the polarity of a single GPO.

        Args:
            gpo_id: the GPO to get the polarity.
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        Returns:
            Polarity of the specified GPI.

        """
        gpo_polarity = int(
            self.mc.communication.get_register(
                self.GPIO_OUT_POLARITY_REGISTER, servo=servo, axis=axis
            )
        )
        gpo_bit_polarity = self.__get_gpio_bit_value(gpo_polarity, gpo_id)
        return GPIOPolarity(int(gpo_bit_polarity))

    def set_gpo_polarity(
        self,
        gpo_id: GPO,
        polarity: GPIOPolarity,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set the polarity of a single GPO.

        Args:
            gpo_id: the GPO polarity to be set.
            polarity: polarity to be set.
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        """
        gpos_actual_polarity = int(
            self.mc.communication.get_register(
                self.GPIO_OUT_POLARITY_REGISTER, servo=servo, axis=axis
            )
        )
        gpos_new_polarity = self.__set_gpio_bit_value(
            gpos_actual_polarity, gpo_id, bool(polarity.value)
        )
        self.mc.communication.set_register(
            self.GPIO_OUT_POLARITY_REGISTER, gpos_new_polarity, servo=servo, axis=axis
        )

    def set_gpo_voltage_level(
        self,
        gpo_id: GPO,
        voltage_level: DigitalVoltageLevel,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """
        Description:
            Set the board voltage level (not the logic level at the uC) of a single GPO.

            This function generates the GPO set point from the actual GPO value and modifies
            the corresponding bit of the desired GPO.

            Finally, it checks that GPOs final value matches with the desired GPOs set point

            .. warning::
                Ensure the polarity is right, if not the output will be wrong.

        Args:
            gpo_id: the GPO to be set.
            voltage_level: For each bit, 0 if voltage level is LOW, 1 if voltage level is HIGH
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        Raise:
            IMException: if the GPOs final value does not match with the desired GPOs set point.
        """
        new_target_value = bool(voltage_level.value)

        gpos_previous_value = int(
            self.mc.communication.get_register(self.GPIO_OUT_VALUE_REGISTER, servo=servo, axis=axis)
        )

        gpos_new_set_point = self.__set_gpio_bit_value(
            gpos_previous_value, gpo_id, new_target_value
        )
        self.mc.communication.set_register(
            self.GPIO_OUT_SET_POINT_REGISTER, gpos_new_set_point, servo=servo, axis=axis
        )
        gpos_final_value = int(
            self.mc.communication.get_register(self.GPIO_OUT_VALUE_REGISTER, servo=servo, axis=axis)
        )

        gpo_final_value = self.__get_gpio_bit_value(gpos_final_value, gpo_id)

        if gpo_final_value != new_target_value:
            raise IMException("Unable to set the GPOs set point value.")



from abc import ABC, abstractmethod
from typing import Optional

from virtual_drive.environment import Environment as VirtualDriveEnvironment


class DriveEnvironmentController(ABC):
    """Abstract Environment Controller.

    Defines methods to control the environment state of the setup,
    which are not controlled by ingeniamotion communications, but are controlled by the user,
    or other hardware.
    """

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def set_gpi(self, number: int, value: bool):
        pass


class ManualUserEnvironmentController(DriveEnvironmentController):
    def __init__(self, pytestconfig):
        self.pytestconfig = pytestconfig
        self.__capsys = (
            None  # Obtain during test execution. Can not be obtain during fixture creation
        )

    @property
    def capsys(self):
        if self.__capsys is None:
            self.__capsys = self.pytestconfig.pluginmanager.getplugin("capturemanager")
        return self.__capsys

    def reset(self):
        pass

    def __request_action_to_user(self, message: str):
        self.capsys.suspend_global_capture(in_=True)
        input(f"{message} [ENTER to confirm]")
        self.capsys.resume_global_capture()

    def set_gpi(self, number: int, value: bool):
        self.__request_action_to_user(f"Please, set gpi {number} to {value}")


class RackServiceEnvironmentController(DriveEnvironmentController):
    """Controller of the environment of Rack Service Setup"""

    def __init__(self, rack_service_client_root, default_drive_idx: Optional[int] = None):
        self.service = rack_service_client_root
        self.default_drive_idx = default_drive_idx

    def reset(self):
        pass

    def set_gpi(self, number: int, value: bool, drive_idx: Optional[int] = None):
        if drive_idx is None:
            drive_idx = self.default_drive_idx
        self.service.write_drive_gpio(drive_idx, f"GPI_{number}", int(value))


class VirtualDriveEnvironmentController(DriveEnvironmentController):
    """Controller of the environment of a Virtual Drive Setup"""

    def __init__(self, virtual_drive_environment: VirtualDriveEnvironment):
        self.__env = virtual_drive_environment
        self.reset()

    def set_gpi(self, number: int, value: bool):
        if number == 1:
            self.__env.gpi_1_status.set(value)
        elif number == 2:
            self.__env.gpi_2_status.set(value)
        elif number == 3:
            self.__env.gpi_3_status.set(value)
        elif number == 4:
            self.__env.gpi_4_status.set(value)
        else:
            raise ValueError

    def reset(self):
        self.__env.gpi_1_status.set(False)
        self.__env.gpi_2_status.set(False)
        self.__env.gpi_3_status.set(False)
        self.__env.gpi_4_status.set(False)



import pytest

from ingeniamotion import MotionController
from ingeniamotion.capture import Capture
from ingeniamotion.communication import Communication
from ingeniamotion.configuration import Configuration
from ingeniamotion.drive_tests import DriveTests
from ingeniamotion.enums import OperationMode
from ingeniamotion.errors import Errors
from ingeniamotion.exceptions import IMStatusWordError
from ingeniamotion.information import Information
from ingeniamotion.metaclass import MCMetaClass
from ingeniamotion.motion import Motion


@pytest.mark.virtual
@pytest.mark.smoke
def test_motion_controller():
    mc = MotionController()
    assert isinstance(mc.configuration, Configuration)
    assert isinstance(mc.motion, Motion)
    assert isinstance(mc.capture, Capture)
    assert isinstance(mc.communication, Communication)
    assert isinstance(mc.tests, DriveTests)
    assert isinstance(mc.errors, Errors)
    assert isinstance(mc.info, Information)


@pytest.mark.virtual
@pytest.mark.smoke
def test_servo_name(motion_controller):
    mc, alias, environment = motion_controller
    prod_code = mc.servos[alias].info["product_code"]
    servo_arg = () if alias == "default" else (alias,)
    name = mc.servo_name(*servo_arg)
    assert name == "{} ({})".format(prod_code, alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_register_enum(motion_controller):
    mc, alias, environment = motion_controller
    servo_arg = () if alias == "default" else (alias,)
    operation_mode_enum = mc.get_register_enum("DRV_OP_VALUE", *servo_arg)
    operation_mode_values = [op.value for op in operation_mode_enum]
    checked_ops = 0
    for element in OperationMode:
        if element.value in operation_mode_values:
            test_name = operation_mode_enum(element.value).name.replace("-", " ")
            name = element.name.replace("_", " ")
            assert test_name.upper() == name
            checked_ops += 1

    assert checked_ops > 0


@pytest.mark.virtual
@pytest.mark.smoke
def test_is_alive(motion_controller):
    mc, alias, environment = motion_controller
    assert mc.is_alive(alias)


@pytest.mark.virtual
class TestMetaclass:
    class DummyClass(metaclass=MCMetaClass):
        mc = MotionController()

        def is_motor_enabled(self, servo: str, axis: int):
            return servo == "a" and axis == 1

        def __init__(self):
            self.mc.servos = ["a", "b"]
            self.mc.configuration.is_motor_enabled = self.is_motor_enabled

        @MCMetaClass.check_motor_disabled
        def dummy_func(self, servo: str, axis: int):
            pass

    @pytest.mark.virtual
    @pytest.mark.parametrize(
        "servo, axis, error",
        [
            ("a", 1, IMStatusWordError),
            ("b", 1, None),
            ("c", 1, KeyError),
            ("a", 2, None),
        ],
    )
    def test_check_motor_disabled_keyword_parameters(self, servo, axis, error):
        dummy_class = self.DummyClass()
        if error is None:
            dummy_class.dummy_func(servo=servo, axis=axis)
        else:
            with pytest.raises(error):
                dummy_class.dummy_func(servo=servo, axis=axis)

    @pytest.mark.virtual
    @pytest.mark.parametrize(
        "servo, axis, error",
        [
            ("a", 1, IMStatusWordError),
            ("b", 1, None),
            ("c", 1, KeyError),
            ("a", 2, None),
        ],
    )
    def test_check_motor_disabled_positional_parameters(self, servo, axis, error):
        dummy_class = self.DummyClass()
        if error is None:
            dummy_class.dummy_func(servo, axis)
        else:
            with pytest.raises(error):
                dummy_class.dummy_func(servo, axis)



import typing
from functools import wraps
from queue import Empty, Full, Queue
from typing import Callable


class StopException(Exception):
    """Stop exception."""


T = typing.TypeVar("T")


class Stoppable:
    stop_queue: Queue[StopException] = Queue(1)

    @staticmethod
    def stoppable(fun: Callable[..., T]) -> Callable[..., T]:
        @wraps(fun)
        def wrapper(self, *args, **kwargs):  # type: ignore
            self.check_stop()
            return fun(self, *args, **kwargs)

        return wrapper

    def reset_stop(self) -> None:
        try:
            self.stop_queue.get(block=False)
        except Empty:
            pass

    def stop(self) -> None:
        try:
            self.stop_queue.put(StopException(), block=False)
        except Full:
            pass

    def check_stop(self) -> None:
        try:
            stop_exception = self.stop_queue.get(block=False)
        except Empty:
            pass
        else:
            raise stop_exception

    def stoppable_sleep(self, timeout: float) -> None:
        try:
            stop_exception = self.stop_queue.get(timeout=timeout)
        except Empty:
            pass
        else:
            raise stop_exception



import random
import time

import pytest
from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.exceptions import ILWrongWorkingCount
from ingenialink.pdo import RPDOMap, RPDOMapItem, TPDOMap, TPDOMapItem
from packaging import version

from ingeniamotion.enums import COMMUNICATION_TYPE, OperationMode
from ingeniamotion.exceptions import IMException

from .setups.descriptors import EthercatMultiSlaveSetup


@pytest.mark.soem
def test_create_rpdo_item(motion_controller):
    mc, alias, environment = motion_controller
    position_set_point_initial_value = 100
    position_set_point = mc.capture.pdo.create_pdo_item(
        "CL_POS_SET_POINT_VALUE", servo=alias, value=position_set_point_initial_value
    )
    assert isinstance(position_set_point, RPDOMapItem)
    assert position_set_point.value == position_set_point_initial_value


@pytest.mark.soem
def test_create_tpdo_item(motion_controller):
    mc, alias, environment = motion_controller
    actual_position = mc.capture.pdo.create_pdo_item("CL_POS_FBK_VALUE", servo=alias)
    assert isinstance(actual_position, TPDOMapItem)


@pytest.mark.soem
def test_create_rpdo_item_no_initial_value(motion_controller):
    mc, alias, environment = motion_controller
    with pytest.raises(AttributeError):
        mc.capture.pdo.create_pdo_item("CL_POS_SET_POINT_VALUE", servo=alias)


@pytest.mark.soem
def test_create_empty_rpdo_map(motion_controller):
    mc, alias, environment = motion_controller
    rpdo_map = mc.capture.pdo.create_empty_rpdo_map()
    assert isinstance(rpdo_map, RPDOMap)
    assert len(rpdo_map.items) == 0


@pytest.mark.soem
def test_create_empty_tpdo_map(motion_controller):
    mc, alias, environment = motion_controller
    tpdo_map = mc.capture.pdo.create_empty_tpdo_map()
    assert isinstance(tpdo_map, TPDOMap)
    assert len(tpdo_map.items) == 0


@pytest.mark.soem
def test_create_pdo_maps_single_item(motion_controller):
    mc, alias, environment = motion_controller
    position_set_point = mc.capture.pdo.create_pdo_item(
        "CL_POS_SET_POINT_VALUE", servo=alias, value=0
    )
    actual_position = mc.capture.pdo.create_pdo_item("CL_POS_FBK_VALUE", servo=alias)
    rpdo_map, tpdo_map = mc.capture.pdo.create_pdo_maps(position_set_point, actual_position)
    assert isinstance(rpdo_map, RPDOMap)
    assert len(rpdo_map.items) == 1
    assert position_set_point in rpdo_map.items
    assert isinstance(tpdo_map, TPDOMap)
    assert len(tpdo_map.items) == 1
    assert actual_position in tpdo_map.items


@pytest.mark.soem
def test_create_pdo_maps_list_items(motion_controller):
    mc, alias, environment = motion_controller
    rpdo_regs = ["CL_POS_SET_POINT_VALUE", "CL_VEL_SET_POINT_VALUE"]
    tpdo_regs = ["CL_POS_FBK_VALUE", "CL_VEL_FBK_VALUE"]
    rpdo_items = [
        mc.capture.pdo.create_pdo_item(rpdo_reg, servo=alias, value=0) for rpdo_reg in rpdo_regs
    ]
    tpdo_items = [mc.capture.pdo.create_pdo_item(tpdo_reg, servo=alias) for tpdo_reg in tpdo_regs]
    rpdo_map, tpdo_map = mc.capture.pdo.create_pdo_maps(rpdo_items, tpdo_items)
    assert isinstance(rpdo_map, RPDOMap)
    assert len(rpdo_map.items) == len(rpdo_items)
    assert isinstance(tpdo_map, TPDOMap)
    assert len(tpdo_map.items) == len(tpdo_items)


@pytest.mark.soem
@pytest.mark.parametrize(
    "register, value, pdo_type",
    [
        ("CL_POS_FBK_VALUE", None, "tpdo"),
        ("CL_VEL_SET_POINT_VALUE", 0, "rpdo"),
    ],
)
def test_add_pdo_item_to_map(motion_controller, register, value, pdo_type):
    mc, alias, environment = motion_controller
    if pdo_type == "rpdo":
        pdo_map = mc.capture.pdo.create_empty_rpdo_map()
    else:
        pdo_map = mc.capture.pdo.create_empty_tpdo_map()
    pdo_map_item = mc.capture.pdo.create_pdo_item(register, servo=alias, value=value)
    mc.capture.pdo.add_pdo_item_to_map(pdo_map_item, pdo_map)
    assert len(pdo_map.items) == 1
    assert pdo_map_item in pdo_map.items


@pytest.mark.soem
@pytest.mark.parametrize(
    "register, value, pdo_type",
    [
        ("CL_POS_FBK_VALUE", None, "tpdo"),
        ("CL_VEL_SET_POINT_VALUE", 0, "rpdo"),
    ],
)
def test_add_pdo_item_to_map_exceptions(motion_controller, register, value, pdo_type):
    mc, alias, environment = motion_controller
    if pdo_type == "tpdo":
        pdo_map = mc.capture.pdo.create_empty_rpdo_map()
    else:
        pdo_map = mc.capture.pdo.create_empty_tpdo_map()
    pdo_map_item = mc.capture.pdo.create_pdo_item(register, servo=alias, value=value)
    with pytest.raises(ValueError):
        mc.capture.pdo.add_pdo_item_to_map(pdo_map_item, pdo_map)


@pytest.mark.soem
@pytest.mark.parametrize(
    "rpdo_maps, tpdo_maps",
    [
        [["rpdo", "rpdo"], ["rpdo", "tpdo"]],
        [["rpdo", "tpdo"], ["tpdo", "tpdo"]],
    ],
)
def test_set_pdo_maps_to_slave_exception(motion_controller, rpdo_maps, tpdo_maps):
    mc, alias, environment = motion_controller
    rx_maps = []
    tx_maps = []
    for map_type in rpdo_maps:
        pdo_map = (
            mc.capture.pdo.create_empty_rpdo_map()
            if map_type == "rpdo"
            else mc.capture.pdo.create_empty_tpdo_map()
        )
        rx_maps.append(pdo_map)
    for map_type in tpdo_maps:
        pdo_map = (
            mc.capture.pdo.create_empty_rpdo_map()
            if map_type == "rpdo"
            else mc.capture.pdo.create_empty_tpdo_map()
        )
        tx_maps.append(pdo_map)
    with pytest.raises(ValueError):
        mc.capture.pdo.set_pdo_maps_to_slave(rx_maps, tx_maps, alias)


@pytest.mark.parametrize(
    "refresh_rate",
    [
        0.0001,
        5,
    ],
)
@pytest.mark.soem
def test_pdos_refresh_rate(motion_controller, refresh_rate):
    mc, alias, environment = motion_controller
    with pytest.raises(ValueError):
        mc.capture.pdo.start_pdos(COMMUNICATION_TYPE.Ethercat, refresh_rate)


@pytest.mark.soem_multislave
@pytest.mark.smoke
def test_start_pdos(motion_controller, tests_setup):
    if not isinstance(tests_setup, EthercatMultiSlaveSetup):
        raise ValueError("Invalid setup config for test")

    mc, aliases, environment = motion_controller

    pdo_map_items = {}
    initial_operation_modes = {}
    rpdo_values = {}
    tpdo_values = {}
    for alias in aliases:
        rpdo_map = mc.capture.pdo.create_empty_rpdo_map()
        tpdo_map = mc.capture.pdo.create_empty_tpdo_map()
        initial_operation_mode = mc.motion.get_operation_mode(servo=alias)
        operation_mode = mc.capture.pdo.create_pdo_item(
            "DRV_OP_CMD", servo=alias, value=initial_operation_mode.value
        )
        actual_position = mc.capture.pdo.create_pdo_item("CL_POS_FBK_VALUE", servo=alias)
        mc.capture.pdo.add_pdo_item_to_map(operation_mode, rpdo_map)
        mc.capture.pdo.add_pdo_item_to_map(actual_position, tpdo_map)
        mc.capture.pdo.set_pdo_maps_to_slave(rpdo_map, tpdo_map, servo=alias)
        pdo_map_items[alias] = (operation_mode, actual_position)
        random_op_mode = random.choice(
            [op_mode for op_mode in OperationMode if op_mode != initial_operation_mode]
        )
        initial_operation_modes[alias] = initial_operation_mode
        rpdo_values[alias] = random_op_mode

    def send_callback():
        for alias in aliases:
            rpdo_map_item, _ = pdo_map_items[alias]
            rpdo_map_item.value = rpdo_values[alias].value

    def receive_callback():
        for alias in aliases:
            _, tpdo_map_item = pdo_map_items[alias]
            tpdo_values[alias] = tpdo_map_item.value

    mc.capture.pdo.subscribe_to_send_process_data(send_callback)
    mc.capture.pdo.subscribe_to_receive_process_data(receive_callback)
    assert not mc.capture.pdo.is_active
    refresh_rate = 0.5
    mc.capture.pdo.start_pdos(refresh_rate=refresh_rate)
    assert mc.capture.pdo.is_active
    time.sleep(2 * refresh_rate)
    mc.capture.pdo.stop_pdos()
    assert not mc.capture.pdo.is_active
    for alias in aliases:
        # Check that RPDO are being sent
        assert rpdo_values[alias] == mc.motion.get_operation_mode(servo=alias)
        # Check that TPDO are being received
        assert pytest.approx(tpdo_values[alias], abs=2) == mc.motion.get_actual_position(
            servo=alias
        )
        # Restore the initial operation mode
        mc.motion.set_operation_mode(initial_operation_modes[alias], servo=alias)
        mc.capture.pdo.remove_rpdo_map(alias, rpdo_map_index=0)
        mc.capture.pdo.remove_tpdo_map(alias, tpdo_map_index=0)


@pytest.mark.soem
def test_stop_pdos_exception(motion_controller):
    mc, alias, environment = motion_controller
    with pytest.raises(IMException):
        mc.capture.pdo.stop_pdos()


@pytest.mark.soem
def test_start_pdos_not_implemented_exception(motion_controller):
    mc, alias, environment = motion_controller
    with pytest.raises(NotImplementedError):
        mc.capture.pdo.start_pdos(COMMUNICATION_TYPE.Canopen)


@pytest.mark.soem
def test_start_pdos_wrong_network_type_exception(motion_controller):
    mc, alias, environment = motion_controller
    with pytest.raises(ValueError):
        mc.capture.pdo.start_pdos("ethernet")


@pytest.mark.soem
def test_start_pdos_number_of_network_exception(mocker, motion_controller):
    mc, alias, environment = motion_controller
    mock_net = {"ifname1": EthercatNetwork("ifname1"), "ifname2": EthercatNetwork("ifname2")}
    mocker.patch.object(mc, "_MotionController__net", mock_net)
    with pytest.raises(ValueError):
        mc.capture.pdo.start_pdos()
    with pytest.raises(IMException):
        mc.capture.pdo.start_pdos(COMMUNICATION_TYPE.Ethercat)


def skip_if_pdo_padding_is_not_available(mc, alias):
    # Check if monitoring is available (To discard EVE-XCR-E)
    try:
        mc.capture._check_version(alias)
    except NotImplementedError:
        is_monitoring_available = False
    else:
        is_monitoring_available = True
    pdo_poller_fw_version = "2.5.0"
    firmware_version = mc.configuration.get_fw_version(alias, 1)
    if (
        version.parse(firmware_version) < version.parse(pdo_poller_fw_version)
        or not is_monitoring_available
    ):
        pytest.skip(
            f"PDO poller is available for firmware version {pdo_poller_fw_version} or higher. "
            f"Firmware version found: {firmware_version}"
        )


@pytest.mark.soem
def test_create_poller(motion_controller):
    mc, alias, environment = motion_controller
    skip_if_pdo_padding_is_not_available(mc, alias)
    registers = [{"name": "CL_POS_FBK_VALUE", "axis": 1}, {"name": "CL_VEL_FBK_VALUE", "axis": 1}]
    sampling_time = 0.25
    samples_target = 4
    sleep_time = (samples_target - 0.5) * sampling_time
    poller = mc.capture.pdo.create_poller(registers, alias, sampling_time)
    time.sleep(sleep_time)
    poller.stop()
    timestamps, data = poller.data
    channel_0_data, channel_1_data = data
    assert len(channel_0_data) == samples_target
    assert len(channel_1_data) == samples_target
    assert len(data) == len(registers)
    assert len(timestamps) == len(channel_0_data)


@pytest.mark.smoke
@pytest.mark.soem
def test_subscribe_exceptions(motion_controller, mocker):
    mc, _, _ = motion_controller

    error_msg = "Test error"

    def start_pdos(self, *args):
        raise ILWrongWorkingCount(error_msg)

    mocker.patch("ingenialink.ethercat.network.EthercatNetwork.stop_pdos")
    mocker.patch(
        "ingenialink.ethercat.network.EthercatNetwork.start_pdos",
        new=start_pdos,
    )
    patch_callback = mocker.patch("ingeniamotion.pdo.PDONetworkManager._notify_exceptions")

    mc.capture.pdo.subscribe_to_exceptions(patch_callback)
    mc.capture.pdo.start_pdos()

    t = time.time()
    timeout = 1
    while not mc.capture.pdo._pdo_thread._pd_thread_stop_event.is_set() and (
        (time.time() - t) < timeout
    ):
        pass

    assert mc.capture.pdo._pdo_thread._pd_thread_stop_event.is_set()
    patch_callback.assert_called_once()
    assert (
        str(patch_callback.call_args_list[0][0][0])
        == f"Stopping the PDO thread due to the following exception: {error_msg} "
    )
    mc.capture.pdo.stop_pdos()



from .descriptors import (
    DriveCanOpenSetup,
    DriveEcatSetup,
    DriveEthernetSetup,
    EthercatMultiSlaveSetup,
)

ETH_EVE_SETUP = DriveEthernetSetup(
    dictionary="//azr-srv-ingfs1/pool/distext/products/EVE-XCR/firmware/2.4.0/eve-xcr-c_eth_2.4.0.xdf",
    ip="192.168.2.10",
    identifier="eve-xcr-c",
    config_file="//azr-srv-ingfs1/dist/setups/setup_eve_can/1.2.0/config.xml",
    fw_file="//azr-srv-ingfs1/pool/distext/products/EVE-XCR/firmware/2.4.0/eve-xcr-c_2.4.0.sfu",
    use_rack_service=True,
)

ETH_CAP_SETUP = DriveEthernetSetup(
    dictionary="//azr-srv-ingfs1/pool/distext/products/CAP-XCR/firmware/2.4.0/cap-xcr-c_eth_2.4.0.xdf",
    ip="192.168.2.11",
    identifier="cap-xcr-c",
    config_file="//azr-srv-ingfs1/dist/setups/setup_cap_can/1.1.0/config.xml",
    fw_file="//azr-srv-ingfs1/pool/distext/products/CAP-XCR/firmware/2.4.0/cap-xcr-c_2.4.0.lfu",
    use_rack_service=True,
)

ECAT_EVE_SETUP = DriveEcatSetup(
    dictionary="//azr-srv-ingfs1/pool/distext/products/EVE-XCR/firmware/2.5.1/eve-xcr-e_eoe_2.5.1.xdf",
    identifier="eve-xcr-e",
    config_file="//azr-srv-ingfs1/dist/setups/setup_eve_ecat/1.2.0/config.xml",
    fw_file="//azr-srv-ingfs1/pool/distext/products/EVE-XCR/firmware/2.5.1/eve-xcr-e_2.5.1.sfu",
    ifname="\\Device\\NPF_{B24AA996-414A-4F95-95E6-2828D346209A}",
    slave=1,
    eoe_comm=True,
    boot_in_app=True,
    use_rack_service=True,
)

ECAT_CAP_SETUP = DriveEcatSetup(
    dictionary="//azr-srv-ingfs1/pool/distext/products/CAP-XCR/firmware/2.5.1/cap-xcr-e_eoe_2.5.1.xdf",
    identifier="cap-xcr-e",
    config_file="//azr-srv-ingfs1/dist/setups/setup_cap_ecat/1.1.0/config.xml",
    fw_file="//azr-srv-ingfs1/pool/distext/products/CAP-XCR/firmware/2.5.1/cap-xcr-e_2.5.1.lfu",
    ifname="\\Device\\NPF_{B24AA996-414A-4F95-95E6-2828D346209A}",
    slave=2,
    eoe_comm=True,
    boot_in_app=False,
    use_rack_service=True,
)

CAN_EVE_SETUP = DriveCanOpenSetup(
    dictionary="//azr-srv-ingfs1/pool/distext/products/EVE-XCR/firmware/2.4.0/eve-xcr-c_can_2.4.0.xdf",
    identifier="eve-xcr-c",
    config_file="//azr-srv-ingfs1/dist/setups/setup_eve_can/1.2.0/config.xml",
    fw_file="//azr-srv-ingfs1/pool/distext/products/EVE-XCR/firmware/2.4.0/eve-xcr-c_2.4.0.sfu",
    device="pcan",
    channel=0,
    node_id=20,
    baudrate=1000000,
    use_rack_service=True,
)

CAN_CAP_SETUP = DriveCanOpenSetup(
    dictionary="//azr-srv-ingfs1/pool/distext/products/CAP-XCR/firmware/2.4.0/cap-xcr-c_can_2.4.0.xdf",
    identifier="cap-xcr-c",
    config_file="//azr-srv-ingfs1/dist/setups/setup_cap_can/1.1.0/config.xml",
    fw_file="//azr-srv-ingfs1/pool/distext/products/CAP-XCR/firmware/2.4.0/cap-xcr-c_2.4.0.lfu",
    device="pcan",
    channel=0,
    node_id=21,
    baudrate=1000000,
    use_rack_service=True,
)

ECAT_MULTISLAVE_SETUP = EthercatMultiSlaveSetup([ECAT_EVE_SETUP, ECAT_CAP_SETUP])



import functools
from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class Setup:
    """Generic setup"""

    pass


@dataclass(frozen=True)
class EthernetSetup(Setup):
    """Any setup that uses Ethernet"""

    ip: str


@dataclass(frozen=True)
class VirtualDriveSetup(EthernetSetup):
    """Setup with virtual drive"""

    dictionary: str
    port: int


@dataclass(frozen=True)
class DriveHwSetup(Setup):
    """Setup with physical hw drive"""

    dictionary: str
    identifier: str
    config_file: Optional[str]
    fw_file: str
    use_rack_service: bool

    @functools.lru_cache()
    def get_rack_drive(self, rack_service_client):
        config = rack_service_client.exposed_get_configuration()
        for idx, drive in enumerate(config.drives):
            if self.identifier == drive.identifier:
                return idx, drive

        raise ValueError(
            f"The drive {self.identifier} cannot be found on the rack's configuration."
        )


@dataclass(frozen=True)
class DriveEthernetSetup(DriveHwSetup, EthernetSetup):
    """Setup with drive with Ethernet.

    Can be regular Ethernet or ethernet over Ethercat (EoE)
    """

    pass


@dataclass(frozen=True)
class DriveEcatSetup(DriveHwSetup):
    """Setup with drive connected with Ethercat"""

    ifname: str
    slave: int
    eoe_comm: bool
    boot_in_app: bool


@dataclass(frozen=True)
class DriveCanOpenSetup(DriveHwSetup):
    """Setup with drive connected with canopen"""

    device: str
    channel: int
    node_id: int
    baudrate: int


@dataclass(frozen=True)
class EthercatMultiSlaveSetup(Setup):
    """Setup with multiple drives connected with Ethercat"""

    drives: list[DriveEcatSetup]



import os
import platform
import tempfile
import time
from collections import OrderedDict
from dataclasses import dataclass

import pytest
from ingenialink.canopen.network import CAN_BAUDRATE, CAN_DEVICE, CanopenNetwork
from ingenialink.canopen.servo import CanopenServo
from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.ethernet.network import EthernetNetwork
from ingenialink.exceptions import ILError
from ingenialink.network import SlaveInfo
from ingenialink.servo import SERVO_STATE

import ingeniamotion
from ingeniamotion import MotionController
from ingeniamotion.exceptions import (
    IMFirmwareLoadError,
    IMRegisterNotExist,
    IMRegisterWrongAccess,
)

from .setups.descriptors import DriveCanOpenSetup, EthernetSetup, Setup

TEST_ENSEMBLE_FW_FILE = "tests/resources/example_ensemble_fw.zfu"


class RegisterUpdateTest:
    def __init__(self):
        self.call_count = 0
        self.alias = None
        self.servo = None
        self.register = None
        self.value = None

    def register_update_test(self, alias, servo, register, value):
        self.call_count += 1
        self.alias = alias
        self.servo = servo
        self.register = register
        self.value = value


class EmcyTest:
    def __init__(self):
        self.messages = []

    def emcy_callback(self, alias, emcy_msg):
        self.messages.append((alias, emcy_msg))


@pytest.mark.virtual
def test_connect_servo_eoe(tests_setup: EthernetSetup):
    mc = MotionController()
    assert "ethernet_test" not in mc.servos
    assert "ethernet_test" not in mc.net
    mc.communication.connect_servo_eoe(
        tests_setup.ip, tests_setup.dictionary, alias="ethernet_test"
    )
    assert "ethernet_test" in mc.servos and mc.servos["ethernet_test"] is not None
    assert "ethernet_test" in mc.net and mc.net["ethernet_test"] is not None


@pytest.mark.virtual
def test_connect_servo_eoe_no_dictionary_error(tests_setup: EthernetSetup):
    mc = MotionController()
    with pytest.raises(FileNotFoundError):
        mc.communication.connect_servo_eoe(tests_setup.ip, "no_dictionary", alias="ethernet_test")


@pytest.mark.virtual
def test_connect_servo_ethernet(tests_setup: EthernetSetup):
    mc = MotionController()
    assert "ethernet_test" not in mc.servos
    assert "ethernet_test" not in mc.net
    mc.communication.connect_servo_ethernet(
        tests_setup.ip, tests_setup.dictionary, alias="ethernet_test"
    )
    assert "ethernet_test" in mc.servos and mc.servos["ethernet_test"] is not None
    assert "ethernet_test" in mc.net and mc.net["ethernet_test"] is not None


@pytest.mark.virtual
def test_connect_servo_ethernet_no_dictionary_error(tests_setup: EthernetSetup):
    mc = MotionController()
    with pytest.raises(FileNotFoundError):
        mc.communication.connect_servo_ethernet(
            tests_setup.ip, "no_dictionary", alias="ethernet_test"
        )


@pytest.mark.smoke
@pytest.mark.ethernet
@pytest.mark.parametrize(
    "coco_dict_path",
    [
        True,
        False,
    ],
)
def test_connect_servo_comkit_no_dictionary_error(coco_dict_path, tests_setup: EthernetSetup):
    mc = MotionController()
    if coco_dict_path:
        coco_dict_path = tests_setup.dictionary
        moco_dict_path = "no_dictionary"
    else:
        coco_dict_path = "no_dictionary"
        moco_dict_path = tests_setup.dictionary
    with pytest.raises(FileNotFoundError):
        mc.communication.connect_servo_comkit(
            tests_setup.ip, coco_dict_path, moco_dict_path, alias="ethernet_test"
        )


@pytest.mark.smoke
@pytest.mark.virtual
def test_get_ifname_from_interface_ip(mocker):
    class MockAdapter:
        @dataclass
        class IP:
            ip = "192.168.2.1"
            is_IPv4 = True

        def __init__(self, interface_name):
            self.name = interface_name
            self.nice_name = interface_name
            self.ips = [self.IP()]
            self.index = 1

    if platform.system() == "Linux":
        name = "eth0"
    else:
        name = b"{192D1D2F-C684-467D-A637-EC07BD434A63}"
    mock_adapter = MockAdapter(name)
    mocker.patch("ifaddr.get_adapters", return_value=[mock_adapter])
    mc = MotionController()
    ifname = mc.communication.get_ifname_from_interface_ip("192.168.2.1")
    if platform.system() == "Windows":
        assert ifname == "\\Device\\NPF_{192D1D2F-C684-467D-A637-EC07BD434A63}"
    else:
        assert ifname == name


@pytest.mark.smoke
@pytest.mark.virtual
def test_get_ifname_by_index():
    mc = MotionController()
    interface_name_list = mc.communication.get_interface_name_list()
    assert len(interface_name_list) > 0
    for index, interface_name in enumerate(interface_name_list):
        ifname = mc.communication.get_ifname_by_index(index)
        assert isinstance(ifname, str)
        if platform.system() == "Linux":
            assert ifname == interface_name


@pytest.mark.skip(reason='This test enters in conflict with "disable_motor_fixture"')
@pytest.mark.smoke
@pytest.mark.canopen
def test_connect_servo_canopen(tests_setup: DriveCanOpenSetup):
    mc = MotionController()
    assert "canopen_test" not in mc.servos
    assert "canopen_test" not in mc.net
    device = CAN_DEVICE(tests_setup.device)
    baudrate = CAN_BAUDRATE(tests_setup.baudrate)
    mc.communication.connect_servo_canopen(
        device,
        tests_setup.dictionary,
        tests_setup.node_id,
        baudrate,
        tests_setup.channel,
        alias="canopen_test",
    )
    assert "canopen_test" in mc.servos and mc.servos["canopen_test"] is not None
    assert "canopen_test" in mc.net and mc.net["canopen_test"] is not None
    mc.net["canopen_test"].disconnect()


@pytest.mark.smoke
@pytest.mark.canopen
@pytest.mark.skip
def test_connect_servo_canopen_busy_drive_error(motion_controller, tests_setup: DriveCanOpenSetup):
    mc, alias, environment = motion_controller
    assert "canopen_test" not in mc.servos
    assert "canopen_test" not in mc.servo_net
    assert alias in mc.servos
    assert alias in mc.servo_net
    assert mc.servo_net[alias] in mc.net
    device = CAN_DEVICE(tests_setup.device)
    baudrate = CAN_BAUDRATE(tests_setup.baudrate)
    with pytest.raises(ILError):
        mc.communication.connect_servo_canopen(
            device,
            tests_setup.dictionary,
            tests_setup.node_id,
            baudrate,
            tests_setup.channel,
            alias="canopen_test",
        )


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "uid, value",
    [
        ("CL_VOL_Q_SET_POINT", 0.34),
        ("CL_POS_SET_POINT_VALUE", -923),
        ("PROF_POS_OPTION_CODE", 1),
    ],
)
def test_get_register(motion_controller, uid, value):
    mc, alias, environment = motion_controller
    drive = mc.servos[alias]
    drive.write(uid, value)
    test_value = mc.communication.get_register(uid, servo=alias)
    assert pytest.approx(test_value) == value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_register_wrong_uid(motion_controller):
    mc, alias, environment = motion_controller
    with pytest.raises(IMRegisterNotExist):
        mc.communication.get_register("WRONG_UID", servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "uid, value",
    [
        ("CL_VOL_Q_SET_POINT", -234),
        ("CL_POS_SET_POINT_VALUE", 23),
        ("PROF_POS_OPTION_CODE", 54),
    ],
)
def test_set_register(motion_controller, uid, value):
    mc, alias, environment = motion_controller
    drive = mc.servos[alias]
    mc.communication.set_register(uid, value, servo=alias)
    test_value = drive.read(uid)
    assert pytest.approx(test_value) == value


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_register_wrong_uid(motion_controller):
    mc, alias, environment = motion_controller
    with pytest.raises(IMRegisterNotExist):
        mc.communication.set_register("WRONG_UID", 2, servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "uid, value, fail",
    [
        ("CL_VOL_Q_SET_POINT", -234, False),
        ("CL_VOL_Q_SET_POINT", "I'm not a number", True),
        ("CL_VOL_Q_SET_POINT", 234.4, False),
        ("CL_POS_SET_POINT_VALUE", 1245, False),
        ("CL_POS_SET_POINT_VALUE", -1245, False),
        ("CL_POS_SET_POINT_VALUE", 1245.5421, True),
        ("PROF_POS_OPTION_CODE", -54, True),
        ("PROF_POS_OPTION_CODE", 54, False),
        ("PROF_POS_OPTION_CODE", "54", True),
    ],
)
def test_set_register_wrong_value_type(motion_controller, uid, value, fail):
    mc, alias, environment = motion_controller
    if fail:
        with pytest.raises(TypeError):
            mc.communication.set_register(uid, value, servo=alias)
    else:
        mc.communication.set_register(uid, value, servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_register_wrong_access(motion_controller):
    mc, alias, environment = motion_controller
    uid = "DRV_STATE_STATUS"
    value = 0
    with pytest.raises(IMRegisterWrongAccess):
        mc.communication.set_register(uid, value, servo=alias)


def dummy_callback(status, _, axis):
    pass


@pytest.mark.ethernet
@pytest.mark.soem
@pytest.mark.canopen
@pytest.mark.smoke
def test_subscribe_servo_status(mocker, motion_controller):
    mc, alias, environment = motion_controller
    axis = 1
    patch_callback = mocker.patch("tests.test_communication.dummy_callback")
    mc.communication.subscribe_servo_status(patch_callback, alias)
    time.sleep(0.5)
    mc.motion.motor_enable(alias, axis)
    time.sleep(0.5)
    mc.motion.motor_disable(alias, axis)
    time.sleep(0.5)
    expected_status = [SERVO_STATE.RDY, SERVO_STATE.ENABLED, SERVO_STATE.DISABLED]
    for index, call in enumerate(patch_callback.call_args_list):
        assert call[0][0] == expected_status[index]
        assert call[0][2] == axis


@pytest.mark.virtual
def test_load_firmware_canopen_exception(motion_controller):
    mc, alias, environment = motion_controller
    with pytest.raises(ValueError):
        mc.communication.load_firmware_canopen("fake_fw_file.lfu", servo=alias)


@pytest.mark.virtual
def test_boot_mode_and_load_firmware_ethernet_exception(mocker, motion_controller):
    mc, alias, environment = motion_controller

    mocker.patch.object(mc, "_get_network", return_value=object())
    with pytest.raises(ValueError):
        mc.communication.boot_mode_and_load_firmware_ethernet("fake_fw_file.lfu", servo=alias)


@pytest.mark.virtual
def test_load_firmware_moco_exception(mocker, motion_controller):
    mc, alias, environment = motion_controller
    mocker.patch.object(mc, "_get_network", return_value=object())
    with pytest.raises(ValueError):
        mc.communication.load_firmware_moco("fake_fw_file.lfu", servo=alias)


@pytest.mark.virtual
def test_connect_servo_virtual():
    mc = MotionController()
    mc.communication.connect_servo_virtual(port=1062)
    assert mc.communication._Communication__virtual_drive is not None
    mc.communication.disconnect()
    assert mc.communication._Communication__virtual_drive is None


@pytest.mark.virtual
def test_connect_servo_virtual_custom_dictionary(tests_setup: Setup):
    mc = MotionController()
    mc.communication.connect_servo_virtual(dict_path=tests_setup.dictionary, port=1062)
    assert mc.communication._Communication__virtual_drive is not None
    mc.communication.disconnect()
    assert mc.communication._Communication__virtual_drive is None


@pytest.mark.virtual
def test_scan_servos_canopen_with_info(mocker):
    mc = MotionController()
    detected_slaves = OrderedDict({31: SlaveInfo(1234, 123), 32: SlaveInfo(1234, 123)})
    mocker.patch(
        "ingenialink.canopen.network.CanopenNetwork.scan_slaves_info", return_value=detected_slaves
    )
    assert mc.communication.scan_servos_canopen_with_info(CAN_DEVICE.KVASER) == detected_slaves


@pytest.mark.virtual
def test_scan_servos_canopen(mocker):
    mc = MotionController()
    detected_slaves = [31, 32]
    mocker.patch(
        "ingenialink.canopen.network.CanopenNetwork.scan_slaves", return_value=detected_slaves
    )
    assert mc.communication.scan_servos_canopen(CAN_DEVICE.KVASER) == detected_slaves


@pytest.mark.virtual
def test_scan_servos_ethercat_with_info(mocker):
    mc = MotionController()
    detected_slaves = OrderedDict({1: SlaveInfo(1234, 123), 2: SlaveInfo(1234, 123)})
    mocker.patch("ingenialink.ethercat.network.EthercatNetwork.__init__", return_value=None)
    mocker.patch(
        "ingenialink.ethercat.network.EthercatNetwork.scan_slaves_info",
        return_value=detected_slaves,
    )
    assert mc.communication.scan_servos_ethercat_with_info("") == detected_slaves


@pytest.mark.virtual
def test_scan_servos_ethercat(mocker):
    mc = MotionController()
    detected_slaves = [1, 2]
    mocker.patch("ingenialink.ethercat.network.EthercatNetwork.__init__", return_value=None)
    mocker.patch(
        "ingenialink.ethercat.network.EthercatNetwork.scan_slaves",
        return_value=detected_slaves,
    )
    assert mc.communication.scan_servos_ethercat("") == detected_slaves


@pytest.mark.virtual
def test_unzip_ensemble_fw_file():
    mc = MotionController()
    with tempfile.TemporaryDirectory() as ensemble_temp_dir:
        mapping = mc.communication._Communication__unzip_ensemble_fw_file(
            TEST_ENSEMBLE_FW_FILE, ensemble_temp_dir
        )

        assert mapping == {
            0: (os.path.join(ensemble_temp_dir, "cap-net-1-e_2.4.0.lfu"), 123456, 4660),
            1: (os.path.join(ensemble_temp_dir, "cap-net-2-e_2.4.0.lfu"), 123456, 16781876),
        }


@pytest.mark.virtual
def test__check_ensemble():
    mc = MotionController()
    with tempfile.TemporaryDirectory() as ensemble_temp_dir:
        mapping = mc.communication._Communication__unzip_ensemble_fw_file(
            TEST_ENSEMBLE_FW_FILE, ensemble_temp_dir
        )
    product_code = 123456
    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 4661),
            2: SlaveInfo(product_code, 16781878),
            4: SlaveInfo(product_code, 4662),
            5: SlaveInfo(product_code, 16781879),
            7: SlaveInfo(654321, 1236),
        }
    )
    for slave_id in [1, 2]:
        assert mc.communication._Communication__check_ensemble(slaves, slave_id, mapping) == 1

    for slave_id in [4, 5]:
        assert mc.communication._Communication__check_ensemble(slaves, slave_id, mapping) == 4


@pytest.mark.virtual
def test__check_ensemble_wrong():
    mc = MotionController()
    with tempfile.TemporaryDirectory() as ensemble_temp_dir:
        mapping = mc.communication._Communication__unzip_ensemble_fw_file(
            TEST_ENSEMBLE_FW_FILE, ensemble_temp_dir
        )

    product_code = 123456
    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 4660),
            2: SlaveInfo(product_code, 16781876),
            3: SlaveInfo(654321, 1236),
        }
    )

    with pytest.raises(IMFirmwareLoadError) as exc_info:
        mc.communication._Communication__check_ensemble(slaves, 3, mapping)
    assert (
        str(exc_info.value) == "The firmware file could not be loaded correctly. "
        "The selected drive is not part of the ensemble."
    )

    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 4660),
            2: SlaveInfo(654321, 16781876),
        }
    )
    with pytest.raises(IMFirmwareLoadError) as exc_info:
        mc.communication._Communication__check_ensemble(slaves, 1, mapping)
    assert (
        str(exc_info.value) == "The firmware file could not be loaded correctly. "
        "The slave 2 has wrong product code or revision number."
    )

    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 16781877),
            2: SlaveInfo(product_code, 4661),
        }
    )
    with pytest.raises(IMFirmwareLoadError) as exc_info:
        mc.communication._Communication__check_ensemble(slaves, 2, mapping)
    assert (
        str(exc_info.value) == "The firmware file could not be loaded correctly. "
        "The slave 2 is not detected."
    )


@pytest.mark.virtual
@pytest.mark.parametrize("revision_number,expected_id_offset", [(4660, 0), (16781876, 1)])
def test_check_slave_in_ensemble(revision_number, expected_id_offset):
    mc = MotionController()
    with tempfile.TemporaryDirectory() as ensemble_temp_dir:
        mapping = mc.communication._Communication__unzip_ensemble_fw_file(
            TEST_ENSEMBLE_FW_FILE, ensemble_temp_dir
        )
    product_code = 123456
    slave_info = SlaveInfo(product_code, revision_number)

    slave_id_offset = mc.communication._Communication__check_slave_in_ensemble(slave_info, mapping)

    assert slave_id_offset == expected_id_offset


@pytest.mark.virtual
def test_check_slave_in_ensemble_drive_not_in_ensemble():
    mc = MotionController()
    with tempfile.TemporaryDirectory() as ensemble_temp_dir:
        mapping = mc.communication._Communication__unzip_ensemble_fw_file(
            TEST_ENSEMBLE_FW_FILE, ensemble_temp_dir
        )
    product_code = 654321
    revision_number = 4660
    slave_info = SlaveInfo(product_code, revision_number)

    with pytest.raises(IMFirmwareLoadError) as exc_info:
        mc.communication._Communication__check_slave_in_ensemble(slave_info, mapping)
    assert (
        str(exc_info.value) == "The firmware file could not be loaded correctly. "
        "The selected drive is not part of the ensemble."
    )


@pytest.mark.virtual
def test_load_ensemble_fw_ecat(mocker):
    product_code = 123456
    slaves = OrderedDict(
        {
            1: SlaveInfo(product_code, 4661),
            2: SlaveInfo(product_code, 16781877),
            3: SlaveInfo(product_code, 4662),
            4: SlaveInfo(product_code, 16781878),
            5: SlaveInfo(654321, 1236),
        }
    )

    mc = MotionController()
    mocker.patch("ingenialink.ethercat.network.EthercatNetwork.__init__", return_value=None)
    mocker.patch(
        "ingenialink.ethercat.network.EthercatNetwork.scan_slaves_info", return_value=slaves
    )
    for slave in [1, 2]:
        patch_fw_callback = mocker.patch(
            "ingenialink.ethercat.network.EthercatNetwork.load_firmware"
        )
        mc.communication.load_firmware_ecat("", TEST_ENSEMBLE_FW_FILE, slave=slave)
        assert len(patch_fw_callback.call_args_list) == 2
        assert patch_fw_callback.call_args_list[0][0][0].endswith("cap-net-1-e_2.4.0.lfu")
        assert patch_fw_callback.call_args_list[0][0][1] is False
        assert patch_fw_callback.call_args_list[0][0][2] == 1

        assert patch_fw_callback.call_args_list[1][0][0].endswith("cap-net-2-e_2.4.0.lfu")
        assert patch_fw_callback.call_args_list[1][0][1] is False
        assert patch_fw_callback.call_args_list[1][0][2] == 2

    for slave in [3, 4]:
        patch_fw_callback = mocker.patch(
            "ingenialink.ethercat.network.EthercatNetwork.load_firmware"
        )
        mc.communication.load_firmware_ecat("", TEST_ENSEMBLE_FW_FILE, slave=slave)
        assert len(patch_fw_callback.call_args_list) == 2
        assert patch_fw_callback.call_args_list[0][0][0].endswith("cap-net-1-e_2.4.0.lfu")
        assert patch_fw_callback.call_args_list[0][0][1] is False
        assert patch_fw_callback.call_args_list[0][0][2] == 3

        assert patch_fw_callback.call_args_list[1][0][0].endswith("cap-net-2-e_2.4.0.lfu")
        assert patch_fw_callback.call_args_list[1][0][1] is False
        assert patch_fw_callback.call_args_list[1][0][2] == 4

    with pytest.raises(IMFirmwareLoadError) as exc_info:
        mc.communication.load_firmware_ecat("", TEST_ENSEMBLE_FW_FILE, slave=5)
    assert (
        str(exc_info.value) == "The firmware file could not be loaded correctly. "
        "The selected drive is not part of the ensemble."
    )


@pytest.mark.virtual
def test_load_ensemble_fw_canopen(mocker):
    class MockDictionary:
        def __init__(self) -> None:
            self.path = "path_to_dictionary"

    class MockCanopenServo(CanopenServo):
        def __init__(self, node_id) -> None:
            self.target = node_id
            self._dictionary = MockDictionary()

    servos = {}
    for node_id in range(1, 6):
        servos[str(node_id)] = MockCanopenServo(node_id)

    mc = MotionController()
    net = CanopenNetwork(CAN_DEVICE.KVASER)
    net.servos = list(servos.values())
    mocker.patch("ingeniamotion.motion_controller.MotionController._get_network", return_value=net)
    mocker.patch("ingenialink.canopen.network.CanopenNetwork.connect_to_slave")
    mocker.patch("ingenialink.canopen.network.CanopenNetwork.disconnect_from_slave")
    mc._get_drive = lambda x: servos[x]
    mc.servos = servos

    product_code = 123456
    slaves_info = OrderedDict(
        {
            1: SlaveInfo(product_code, 4661),
            2: SlaveInfo(product_code, 16781877),
            3: SlaveInfo(product_code, 4662),
            4: SlaveInfo(product_code, 16781878),
            5: SlaveInfo(654321, 1236),
        }
    )
    mocker.patch(
        "ingenialink.canopen.network.CanopenNetwork.scan_slaves_info", return_value=slaves_info
    )
    for slave in [1, 2]:
        patch_fw_callback = mocker.patch("ingenialink.canopen.network.CanopenNetwork.load_firmware")
        mc.communication.load_firmware_canopen(TEST_ENSEMBLE_FW_FILE, servo=str(slave))
        assert len(patch_fw_callback.call_args_list) == 2
        assert patch_fw_callback.call_args_list[0][0][0] == 1
        assert patch_fw_callback.call_args_list[0][0][1].endswith("cap-net-1-e_2.4.0.lfu")

        assert patch_fw_callback.call_args_list[1][0][0] == 2
        assert patch_fw_callback.call_args_list[1][0][1].endswith("cap-net-2-e_2.4.0.lfu")

    for slave in [3, 4]:
        patch_fw_callback = mocker.patch("ingenialink.canopen.network.CanopenNetwork.load_firmware")
        mc.communication.load_firmware_canopen(TEST_ENSEMBLE_FW_FILE, servo=str(slave))
        assert len(patch_fw_callback.call_args_list) == 2
        assert patch_fw_callback.call_args_list[0][0][0] == 3
        assert patch_fw_callback.call_args_list[0][0][1].endswith("cap-net-1-e_2.4.0.lfu")

        assert patch_fw_callback.call_args_list[1][0][0] == 4
        assert patch_fw_callback.call_args_list[1][0][1].endswith("cap-net-2-e_2.4.0.lfu")

    with pytest.raises(IMFirmwareLoadError) as exc_info:
        mc.communication.load_firmware_canopen(TEST_ENSEMBLE_FW_FILE, servo="5")
    assert (
        str(exc_info.value) == "The firmware file could not be loaded correctly. "
        "The selected drive is not part of the ensemble."
    )


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "net_types", [[EthernetNetwork, CanopenNetwork], [EthercatNetwork, EthernetNetwork], []]
)
def test_get_available_canopen_devices_check_get_available_devices_call(mocker, net_types):
    mc = ingeniamotion.MotionController()
    test_net = None
    for n, n_type in enumerate(net_types):
        net = mocker.MagicMock(spec=n_type)
        mc.net[n] = net
        if n_type == CanopenNetwork:
            test_net = net
    patch_get_available_devices = mocker.patch(
        "ingenialink.canopen.network.CanopenNetwork.get_available_devices"
    )
    mc.communication.get_available_canopen_devices()
    if test_net is not None:
        assert test_net.get_available_devices.call_count == 1
        assert patch_get_available_devices.call_count == 0
    else:
        assert patch_get_available_devices.call_count == 1


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_available_canopen_devices(mocker):
    mc = ingeniamotion.MotionController()
    mocker.patch(
        "ingenialink.canopen.network.CanopenNetwork.get_available_devices",
        return_value=[
            ("pcan", "PCAN_USBBUS1"),
            ("pcan", "PCAN_USBBUS2"),
            ("kvaser", 0),
            ("kvaser", 1),
            ("ixxat", 0),
            ("ixxat", 1),
        ],
    )
    test_output = mc.communication.get_available_canopen_devices()
    expected_ouput = {CAN_DEVICE.KVASER: [0, 1], CAN_DEVICE.PCAN: [0, 1], CAN_DEVICE.IXXAT: [0, 1]}
    assert test_output == expected_ouput


@pytest.mark.virtual
@pytest.mark.smoke
def test_subscribe_register_updates(motion_controller):
    user_over_voltage_uid = "DRV_PROT_USER_OVER_VOLT"
    register_update_callback = RegisterUpdateTest()

    mc, alias, environment = motion_controller
    mc.communication.subscribe_register_update(
        register_update_callback.register_update_test, servo=alias
    )

    previous_reg_value = mc.communication.get_register(user_over_voltage_uid, servo=alias)
    assert register_update_callback.call_count == 1
    assert register_update_callback.alias == alias
    assert register_update_callback.register.identifier == user_over_voltage_uid
    assert register_update_callback.value == previous_reg_value

    new_reg_value = 100
    mc.communication.set_register(user_over_voltage_uid, value=new_reg_value, servo=alias)
    assert register_update_callback.call_count == 2
    assert register_update_callback.alias == alias
    assert register_update_callback.register.identifier == user_over_voltage_uid
    assert register_update_callback.value == new_reg_value

    mc.communication.unsubscribe_register_update(
        register_update_callback.register_update_test, servo=alias
    )

    mc.communication.set_register(user_over_voltage_uid, value=previous_reg_value, servo=alias)

    # The old value has not been propagated after unsubscribing
    assert register_update_callback.value == new_reg_value


@pytest.mark.canopen
@pytest.mark.soem
@pytest.mark.smoke
def test_emcy_callback(motion_controller):
    mc, alias, _ = motion_controller
    emcy_test = EmcyTest()
    mc.communication.subscribe_emergency_message(emcy_test.emcy_callback, servo=alias)
    prev_val = mc.communication.get_register("DRV_PROT_USER_OVER_VOLT", axis=1, servo=alias)
    mc.communication.set_register("DRV_PROT_USER_OVER_VOLT", value=10.0, axis=1, servo=alias)
    with pytest.raises(ILError):
        mc.motion.motor_enable(servo=alias)
    mc.motion.fault_reset(servo=alias)
    assert len(emcy_test.messages) == 2
    servo_alias, first_emcy = emcy_test.messages[0]
    assert servo_alias == alias
    assert first_emcy.error_code == 0x3231
    assert first_emcy.get_desc() == "User Over-voltage detected"
    servo_alias, second_emcy = emcy_test.messages[1]
    assert servo_alias == alias
    assert second_emcy.error_code == 0x0000
    assert second_emcy.get_desc() == "No error"
    mc.communication.set_register("DRV_PROT_USER_OVER_VOLT", value=prev_val, axis=1, servo=alias)
    mc.communication.unsubscribe_emergency_message(emcy_test.emcy_callback, servo=alias)
